/*
** Module   : Button.java
** Abstract : push button implementation
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050124   @19398 Creation based on CHARVA JButton class.
**                           Necessary modification to get compiled.
** 002 EVL 20050203   @19559 Rollback the #001 approach. Only draw() and
**                           probably processKeyEvent() need to be
**                           replaced.
** 003 EVL 20050207   @19738 Changes to apply SPACEBAR as selector key.
** 004 EVL 20050215   @19829 Adding simulation of Progress behavior for.
**                           focus moving by arrow keys.
** 005 EVL 20050221   @19908 Enhanced moving focus by arrow keys.
** 006 EVL 20050222   @19917 Width dynamic change handling.
** 007 EVL 20050411   @20670 Changing location detection to the
**                           getLocation call instead of
**                           getLocationOnScreen to be able to get the
**                           coordinates relative to the Button's parent
**                           container which can be inside the view-port.
** 008 EVL 20050601   @21373 Changing the inner variable names to sync
**                           with the rest of the project.
** 009 EVL 20050601   @21374 Adding the Progress specific methods.
** 010 EVL 20050602   @21392 Syntax changes to the javadoc comments.
** 011 EVL 20050607   @21435 Syntax for comments checking.
** 012 EVL 20050607   @21438 Removing the visible member. It is completely
**                           duplicated in CHARVA's Component superclass
**                           so we do not need to double this work. Also
**                           the methods to work with the HIDDEN attribute
**                           are modified to reflect this change.
** 013 EVL 20050610   @21483 Rework the getting next/previous siblings for
**                           the button widget.
** 014 EVL 20050614   @21484 Rework the getting next/previous tab item for
**                           the button widget.
** 015 EVL 20050615   @21510 Adoption the attribute to the CHARVA model.
**                           The attributes TAB-POSITION, ROW, COLUMN,
**                           HELP, FRAME-ROW, FRAME-COLUMN are touched.
** 016 EVL 20050729   @21915 Removing the ugly tabulation from source.
** 017 EVL 20050822   @22199 Using the P2JFrame/Container interfaces to
**                           detect the class.
** 018 SIY 20050831   @22487 Refactoring (see @@22394-22409) 
** 019 SIY 20050907   @22624 Fixed use of @inheritDoc tag in javadoc, 
**                           fixed constructor, fixed automatic resizing.
** 020 SIY 20050920   @22826 Added support for SELECTED attribute. 
**                           Using ColorTable for the colors, added
**                           support for PFCOLOR attribute.  
** 021 SIY 20051129   @23557 Commented out call to setParentAttribute().
** 022 SIY 20060131   @24197 Use ThinClient special method instead of
**                           direct calls to Window.message().
** 023 SIY 20060209   @24491 Cleanups and updates to accommodate changes
**                           in xxxConfig classes.
** 024 SIY 20060215   @24595 Refactoring and fixes to make it working 
**                           properly in AlertBox.
** 025 NVS 20060227   @24863 Key actions KA_* are used instead of virtual
**                           keys VK_* for all editing operations.
** 026 SIY 20060515   @26144 Encapsulated access to label and help fields.
** 027 GES 20060803   @28393 Major cleanup and refactoring.
** 028 GES 20060808   @28487 Fixed cursor processing to generate next or
**                           previous focus movement.
** 029 NVS 20060823   @28785 Cursor display is turned off while the button
**                           has focus.
** 030 SIY 20060824   @28836 Added LabeledWidget interface support.
** 031 SIY 20060824   @28842 Removed getP2JType() method.
** 032 NVS 20060824   @28872 Cursor left/right/up/down keys are processed
**                           like in Progress.
** 033 GES 20060825   @28916 Fixed event processing for alert boxes that
**                           was broken in H032.  Fixed drawing of the
**                           mnemonic highlight in a selected button.
** 034 GES 20061117   @31248 Fixed button drawing to properly account for
**                           all possible explicitly set widths. The
**                           getWidth() and getHeight() methods needed
**                           to be implemented to account for explicitly
**                           set width/height such that the layout of
**                           buttons was fixed.
** 035 NVS 20061204   @31543 Implemented TAB and BACK_TAB keys processing
**                           to make the widgets behave uniformly.
** 036 NVS 20061204   @31551 Switched to using a private widgetFocused
**                           flag instead of config.selected
** 037 GES 20061208   @31636 Added safer cursor visibility processing
**                           during focus gain/loss.
** 038 SIY 20061227   @31749 Fixed drawing of focused buttons in some 
**                           special cases.
** 039 NVS 20070426   @33249 Enabled state is checked instead of SENSITIVE
**                           attribute.
** 040 SIY 20070807   @34804 Added isColumnFixed() method.
** 041 SVL 20071221   @36528 Label is shown/hidden together with the
**                           widget. 
** 042 NVS 20080401   @37772 Added isFocusable() method implementation.
** 043 GES 20080421   @38059 I18N charset support.
** 044 NVS 20081226   @40997 Drawing takes into consideration highlighted
**                           state instead of focused. They are not equal
**                           (focused button is not highlighted while out
**                           of wait-for).
** 045 CA  20090115   @41127 Fixed invalid cursor state caused by 
**                           consecutive FOCUS_GAINED events: when the 
**                           button is already marked as 'focused', it
**                           must not overwrite the 'curvis' flag. 
** 046 SIY 20100619          Switching to own implementation of character UI.
** 047 SIY 20100804          Changes in imports, some refactorings.
** 048 SIY 20100822          Refactored AttributeContainer support.
** 049 SIY 20100831          Changes related to refactoring of FrameHelper.
** 050 SIY 20100915          locateFrame() is moved to UiUtils.
** 054 SIY 20101122          Reduced class hierarchy. Switching from KeyEvent 
**                           to KeyInput.
** 055 SIY 20110112          Changes related to new color handling scheme.
** 056 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 057 SIY 20111202          Refactored and moved from CHUI package.
** 058 EVL 20140528          Adding implementation stubs for two new methods.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * An implementation of a "push button" with a text label.
 */
public abstract class Button
extends AbstractWidget
implements KeyListener,
           FocusListener, 
           LabeledWidget
{
   /** Command to generate. */
   private String command;

   /** Button label. */
   private String textLabel;

   /** Container for listeners. */
   private ArrayList<ActionListener> listeners = new ArrayList<ActionListener>();

   /** Button attribute container. */
   private ButtonConfig config = null;

   /** Index of the character to underscore. */
   protected int underscore = -1;
   
   /** Label instance associated with widget. */
   private Label label = null;
   
   /** Widget in focus flag. */
   private boolean widgetFocused = false;
   
   /** Previous cursor visibility flag to restore on focus lost. */
   private boolean curvis = false;
   
   /**
    * This constructor creates a button with a given text label.
    * 
    * @param   owner
    *          The parent container, frame or dialog which will include this
    *          button.
    * @param   text
    *          The string to be used as the button text.
    */
   public Button(Container owner, String text)
   {
      initButtonText(text);
      
      config = new ButtonConfig(ComponentConfig.BUTTON, this);
      
      config.setName(text);
      config.setWidthCharsDirect(width());
      config.setHelp(null);
      addFocusListener(this);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   cfg
    *          Configuration data.
    */
   public Button(ComponentConfig cfg)
   {
      initButtonText("");
      config = new ButtonConfig(this, cfg);
      config.applyConfig(cfg);

      if (cfg.getLabel() != null)
         setText(cfg.getLabel());
      addFocusListener(this);
   }

   /**
    * Initialize button text.
    * 
    * @param   text
    *          Button text.
    */
   protected void initButtonText(String text) 
   {
      setTextLabel(text);
      setCommand(text);
   }

   /**
    * Add action listener.
    * 
    * @param   listener
    *          Listener to add.
    */
   public void addListener(ActionListener listener)
   {
      listeners.add(listener);
   }

   /**
    * Check if passes key does match underscored character in the text.
    * 
    * @param   keyEvent
    *          Event to check.
    * 
    * @return  <code>true</code> if match is found.
    *  
    */
   public boolean checkKey(KeyInput keyEvent)
   {
      if (underscore >= 0 && keyEvent.id() == EventType.KEY_TYPED)
      {
         char ch = Character.toLowerCase(textLabel().charAt(underscore));
         char kc = Utils.toChar(keyEvent.keyCode());
         
         if (ch == Character.toLowerCase(kc))      
            return true;
      }
      
      return false;
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @Override
   @SuppressWarnings("unchecked")
   public ComponentConfig config()
   {
      return config;
   }

   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), height());
   }

   /**
    * Draw the button. Called by this Button's parent container.
    */
   @Override
   public abstract void draw();

   /**
    * Return current value of COLON option.
    * 
    * @return  Current value of the option column.
    */
   @Override
   public int getColon()
   {
      return config.getColon();
   }

   /**
    * Return text of the column label.
    * 
    * @return  text of the column label.
    */
   @Override
   public String getColumnLabelText()
   {
      return null;
   }

   /**
    * Return reference to the instance of the <code>Label</code> attached to
    * the widget.
    * 
    * @return  Reference to <code>Label</code> instance.
    */
   @Override
   public Label getLabelInstance()
   {
      return label;
   }

   /**
    * Return text of the label.
    * 
    * @return  text of the label.
    */
   @Override
   public String getLabelText()
   {
      return null;
   }
   
   /**
    * Return TO column.
    * 
    * @return  Column assigned via TO option.
    */
   @Override
   public int getTo()
   {
      return config.getTo();
   }
   
   /**
    * Calculate the height of this widget.
    *
    * @return    The height in rows.
    */
   @Override
   public int height()
   {
      return config.getHeightChars() < 1 ? 1 : config.getHeightChars();
   }
   
   /**
    * Check if widget has fixed column position.
    * 
    * @return  <code>true</code> if widget column is fixed.
    */
   @Override
   public boolean isColumnFixed()
   {
      return config.getAssignedColumn() != BaseConfig.INV_COORD;
   }

   /**
    * Get current state of flag which is forces presence of the label
    * regardless from the frame NO-LABELS option.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isForceLabel()
   {
      return config.isForceLabel();
   }

   /**
    * Get current state of NO-LABELS flag.
    * 
    * @return  Current state of the flag.
    */
   @Override
   public boolean isNoLabels()
   {
      return config.isNoLabels();
   }

   /**
    * Get widget alignment policy.
    * 
    * @return  <code>true</code> if widget is formatted for right alignment.
    */
   @Override
   public boolean isRightAligned()
   {
      return false;
   }

   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }
   
   /**
    * This is called when the button takes the input focus. When the help
    * string is defined for the button it should be displayed at the window
    * status line.
    *
    * @param   event
    *          The focus event to be processed.
    */
   @Override
   public void onFocusGained(FocusEvent event)
   {
      // force the cursor off since it should not display while we have focus
      boolean state = screen().setCursorStatus(false);
      
      if (!widgetFocused)
      {
         // the cursor state is saved only if the button is not already marked
         // as being focused; else, if the button is already focused, the 
         // cursor state could be different then the state when the first
         // FOCUS_GAINED was triggered, restoring the cursor to an incorrect
         // state, after FOCUS_LOST is triggered.
         curvis = state;
      }

      widgetFocused = true;
      
      // even for consecutive FOCUS_GAINED events, the HELP text must be set
      // for this button, because the HELP message-text may have been changed
      // for another widget (i.e. on a nested UPDATE)
      
      // draw the help string for focused button
      ThinClient.setHelp(config.getHelp());
   }
   
   /**
    * This is called when the button looses the input focus.
    *
    * @param   focusEvent
    *          The focus event to be processed.
    */
   @Override
   public void onFocusLost(FocusEvent focusEvent)
   {
      widgetFocused = false;
      
      // restore the cursor visibility that existed when we gained focus 
      screen().setCursorStatus(curvis);
   }

   /**
    * This method is invoked when key is pressed.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onKeyPressed(KeyInput event)
   {
   }

   /**
    * This method is invoked when key is typed.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void onKeyTyped(KeyInput event)
   {
   }

   /**
    * Process event and dispatch it to specific method.
    * 
    * @param   event
    *          Event to process.
    */
   @Override
   public void processEvent(Event event)
   {
      super.processEvent(event);

      if (event instanceof KeyInput)
      {
         KeyInput keyEvent = (KeyInput) event;
         
         if ((!keyEvent.isConsumed()) &&
             keyEvent.keyCode() == Key.VK_ENTER && isEnabled())
         {
            EventManager.postEvent(new ActionEvent(this, command));
            keyEvent.consume();
         }
         
         return;
      }
      
      if (event instanceof ActionEvent)
      {
         for (ActionListener listener : listeners)
         {
            listener.onAction(((ActionEvent) event));
         }
      }
   }

   /**
    * Processes key events occurring on this object by dispatching them
    * to any registered KeyListener objects.
    *
    * @param    keyEvent
    *           The event to be handled.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      if (handleHotkeys(keyEvent))
         return;
      
      // use arrow keys to move focus between widgets (only left/right work
      // in the alert box case)
      if (parent() instanceof AlertBox)
      {
         setActionCode(keyEvent);
      }
      else
      {
         if (handleCursorKeys(keyEvent.actionCode(), (Frame) UiUtils.locateFrame(this)))
            return;
      }

      handleBasicEvents(keyEvent);
   }

   /**
    * Handle space and keys that match the highlighted char in button text.
    * 
    * @param   ki
    *          Input event.
    *          
    * @return  <code>true</code> if key is handled.
    */
   private boolean handleHotkeys(KeyInput ki)
   {
      boolean returnNow = false;
      int key = ki.actionCode();
      
      // changing the key code to VK_ENTER will cause the button to be
      // "selected" (pressed) downstream
      if (key == ' ')
      {
         ki.setKeyCode(Key.VK_ENTER);
      }
      else if (underscore >= 0 && ki.id() == EventType.KEY_TYPED)
      {
         // test for the mnemonic or accelerator character that is 
         // highlighted on the button
         if (checkKey(ki))
         {
            ki.setKeyCode(Key.VK_ENTER);
         }
         else
         {
            if ((parent() instanceof AlertBox))
            {
               // alert boxes check for mnemonic characters across the
               // entire set of buttons (and if found will manually force
               // the selection of that button)
               if (((AlertBox) parent()).findButton(ki))
                  returnNow  = true;
            }
         }
         
      }
      return returnNow;
   }

   /**
    * Handle cursor movement keys.
    * 
    * @param   key
    *          Key code.
    * @param   frame
    *          Parent frame.
    * 
    * @return  <code>true</code> if any cursor key was found.
    */
   private boolean handleCursorKeys(int key, Frame frame)
   {
      boolean returnNow = false;
      
      if (key == Keyboard.KA_TAB)
      {
         returnNow = nextFocus();
      }
      else if (key == Keyboard.KA_BACK_TAB)
      {
         returnNow = prevFocus();
      }
      else if (key == Keyboard.KA_CURSOR_LEFT)
      {
         // the left most widget ignores this condition
         if (!frame.isLeftMostWidget(this))
         {
            returnNow = prevFocus();
         }
      }
      else if (key == Keyboard.KA_CURSOR_UP)
      {
         returnNow = moveAboveWidget(frame);
      }
      else if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         // the right most widget ignores this condition
         if (!frame.isRightMostWidget(this))
         {
            returnNow = nextFocus();
         }
      }
      else if (key == Keyboard.KA_CURSOR_DOWN)
      {
         returnNow = moveBelowWidget(frame);
      }
      return returnNow;
   }

   /**
    * Move focus to widget below in specified frame.
    * 
    * @param   frame
    *          Parent frame.
    *          
    * @return  <code>true</code>
    */
   private boolean moveBelowWidget(Frame frame)
   {
      frame.moveBelowWidget(this, 0);
      refresh();
      return true;
   }

   /**
    * Move focus to widget above in specified frame.
    * 
    * @param   frame
    *          Parent frame.
    *          
    * @return  <code>true</code>
    */
   private boolean moveAboveWidget(Frame frame)
   {
      frame.moveAboveWidget(this, 0);
      refresh();
      return true;
   }

   /**
    * Move focus to previous widget in Tab order.
    * 
    * @return  <code>true</code>
    */
   private boolean prevFocus()
   {
      ThinClient.previousFocus(this);
      refresh();
      return true;
   }

   /**
    * Move focus to next widget in Tab order.
    * 
    * @return  <code>true</code>
    */
   private boolean nextFocus()
   {
      ThinClient.nextFocus(this);
      refresh();
      return true;
   }

   /**
    * Refresh widget at the screen.
    */
   private void refresh()
   {
      draw();
      requestSync();
   }

   /**
    * Prepare event for processing.
    *   
    * @param   ki
    *          Input event.
    */
   private void setActionCode(KeyInput ki)
   {
      int key = ki.actionCode();
      // alert box processing has no need to generate entry/leave events 
      // so we take a very simple approach
      if (key == Keyboard.KA_CURSOR_LEFT)
      {
         ki.setActionCode(Keyboard.KA_BACK_TAB);
         ki.setKeyCode(Key.VK_BACK_TAB);
      }
      else if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         ki.setActionCode(Keyboard.KA_TAB);
         ki.setKeyCode(Key.VK_TAB);         
      }
   }
   
   /**
    * Process basic events - TAB, BACK-TAB and ENTER key event.
    * 
    * @param   event
    *          Key event to process.
    */
   private void handleBasicEvents(KeyInput event)
   {
      super.processKeyEvent(event);

      if (event.isConsumed())
      {
         return;
      }

      int key = event.keyCode();
      if (key == '\t')
      {
         parent().nextFocus();
         return;
      }
      else if (key == Key.VK_BACK_TAB)
      {
         parent().prevFocus();
         return;
      }

      if (super.isEnabled() && (key == Key.VK_ENTER))
      {
         EventManager.postEvent(new ItemEvent(this));
      }
   }
   
   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      if (isEnabled())
         super.requestFocus();

      screen().at(screenLocation().offset(1,0));
   }

   /**
    * Set new value for the COLON option.
    * 
    * @param   colon
    *          New value for the option column.
    */
   @Override
   public void setColon(int colon)
   {
      config.setColon(colon);
   }
   
   /**
    * Set command associated with button.
    * 
    * @param   text
    *          Command text.
    */
   public void setCommand(String text)
   {
      command = text;
   }

   /**
    * Assign new instance of <code>Label</code> to the widget.
    * 
    * @param   label
    *          Reference to new <code>Label</code> instance.
    */
   @Override
   public void setLabelInstance(Label label)
   {
      this.label = label;
      this.label.setVisible(isVisible());
   }

   /**
    * Force NO-LABELS flag.
    * 
    * @param   labels
    *          New value of the flag.
    */
   @Override
   public void setNoLabels(boolean labels)
   {
      config.setNoLabels(labels);
   }

   /**
    * Set the new text for the component.
    * 
    * @param   text
    *          New text value.
    */
   public void setText(String text)
   {
      setTextLabel(text);

      repaint();
      
      if (config.isAutoResize())
         config.setWidthCharsDirect(textLabel().length() + 2);

   }

   /**
    * Set button label.
    * 
    * @param   label
    *          Label text.
    */
   public void setTextLabel(String label)
   {
      this.textLabel = label;
   }

   /**
    * Assign new TO column.
    * 
    * @param   col
    *          New value for the option.
    */
   @Override
   public void setTo(int col)
   {
      config.setTo(col);
   }

   /**
    * Shows or hides the button and its label.
    *
    * @param    visible
    *           <code>true</code> to make the widget visible, 
    *           <code>false</code> to hide the widget.
    */
   @Override
   public void setVisible(boolean visible)
   {
      super.setVisible(visible);

      if (label != null)
         label.setVisible(visible);
   }
   
   /**
    * Get label text.
    * 
    * @return  label text.
    */
   public String textLabel()
   {
      return textLabel;
   }

   /**
    * Calculate the width of this widget.
    *
    * @return    The width in characters.
    */
   @Override
   public int width()
   {
      boolean max = (config.isAutoResize() || config.getWidthChars() < 1);
      return max ? ((textLabel().length() + 2)) : config.getWidthChars();
   }
   
   /**
    * Set index of the character to underscore.
    * 
    * @param   offset
    *          0-based index of the character to underscore.
    */
   public void setUnderscoreOffset(int offset)
   {
      underscore = offset;
   }
   
   /**
    * Checking if the widget was made visible at least once.
    * 
    * @return  <code>true</code> if the widget was visible,
    *          <code>false</code> otherwise.
    */
   @Override
   public boolean wasVisible()
   {
      return true;
   }
   
   /**
    * Remember the widget was made visible at least once.
    */
   @Override
   public void setWasVisible()
   {
      return;
   }
}
